#include "Automata.h"
#include <cstdlib>
#include <math.h>
#include <assert.h>
#include "Synth.h"

Automata::Automata(int width, int depth)
{
	assert(depth >= 2);

	bufferWidth = width;
	bufferDepth = depth;
	effectiveBufferWidth = bufferWidth;
	rule = 0;
	randomlineChance = 0;
	randomruleChance = 0;

	//Init cell buffer
	cellBuffer = (bool**)calloc(bufferDepth, sizeof(bool*));
	for (int c = 0; c < bufferDepth; c++)
		cellBuffer[c] = (bool*)calloc(bufferWidth, sizeof(bool));

	//Init freqs
	freqMap = (double*)calloc(bufferWidth, sizeof(double));
	for (int c = 0; c < bufferWidth; c++)
	{
		freqMap[c] = 220 * pow(2, c / 12.0);
	}

	//Init volumes
	volMap = (int*)calloc(bufferWidth, sizeof(int));
	for (int c = 0; c < bufferWidth; c++)
	{
		volMap[c] = (SDL_MIX_MAXVOLUME / 2 - 15 * abs(c - bufferWidth / 2)) / bufferWidth;
	}

	freqPeriod = (double*)calloc(bufferWidth, sizeof(double));

	//Seed
	cellBuffer[bufferDepth - 1][bufferWidth / 2] = true;
}


//Set bandwidth that actually maps to notes
void Automata::SetEffectiveWidth(int width)
{
	effectiveBufferWidth = width;
}


//Set the rule to use
void Automata::SetRule(char irule)
{
	rule = irule;
}


//Set the chance for the rule to be randomized
void Automata::SetRandomruleChance(int chance)
{
	randomruleChance = chance;
}


//Set the chance for a line to be randomized
void Automata::SetRandomlineChance(int chance)
{
	randomlineChance = chance;
}


//Manually trigger a line respawn event
void Automata::RandomLine()
{
	//Random line
	for (int c = 0; c < bufferWidth; c++)
		cellBuffer[bufferDepth - 1][c] = rand() % 2 == 0;
}


//Set the frequency of a given column
void Automata::SetFreq(int index, double freq)
{
	freqMap[index] = freq;
}


//Spawn the next iteration
void Automata::Cycle()
{
	//Bump rows down
	bool* row = cellBuffer[0];
	for (int c = bufferDepth - 1; c >= 0; c--)
	{
		bool* temp = row;
		row = cellBuffer[c];
		cellBuffer[c] = temp;
	}

	//Random rule
	if (rand() % 100 < randomruleChance)
		rule = rand() % 256;

	if (rand() % 100 < randomlineChance)
	{
		//Random line
		for (int c = 0; c < bufferWidth; c++)
			cellBuffer[bufferDepth - 1][c] = rand() % 2 == 0;
	}
	else
	{
		//Propagate
		for (int c = 0; c < bufferWidth; c++)
		{
			//Represent previous 3 cells
			int prev = 0;
			for (int d = c-1; d <= c+1; d++)
				if (d > 0 && d < bufferWidth)
					if (cellBuffer[bufferDepth - 2][d])
						prev |= 1<<(c - d + 1);

			cellBuffer[bufferDepth - 1][c] = (rule & 1<<prev) > 0;
		}
	}
}


//Render the sound
void Automata::RenderSines(Uint8* stream, int len, SDL_AudioSpec audioFormat) const
{
	Uint8* sine = (Uint8*)calloc(len, sizeof(Uint8));
	int left = (bufferWidth - effectiveBufferWidth) / 2;
	for (int c = left; c < left + effectiveBufferWidth; c++)
	{
		if (freqMap[c] > 0 && cellBuffer[bufferDepth - 1][c]) //|| freqPeriod[c] != 0)
		{
			for (int d = 0; d < len; d++)
			{
				freqPeriod[c]++;
				double v1 = sin(2 * 3.1416 * freqMap[c] / audioFormat.freq * freqPeriod[c]);
				//if (v1 > 0)
					v1 = (int)(v1 + 0.5);
				//else
				//	v1 = (int)(v1 - 0.5);
				sine[d] = 127 * (1 + v1);

				//Shut down on 0
				/*if (d > 0
					&& !cellBuffer[bufferDepth - 1][c]
					&& ((sine[d] <= 127 && sine[d-1] >= 127) || (sine[d] >= 127 && sine[d-1] <= 127)))
				{
					sine[d] = 0;
					freqPeriod[c] = 0;
					break;
				}*/
			}
			SDL_MixAudio(stream, sine, len, volMap[c]);
		}
		else
			freqPeriod[c] = 0;
	}
	free(sine);
}


void Automata::RenderSynth(Synth* synth, int duration) const
{
	int left = (bufferWidth - effectiveBufferWidth) / 2;
	for (int c = left; c < left + effectiveBufferWidth; c++)
	{
		if (freqMap[c] > 0) synth->PlayTone(freqMap[c], duration);
	}
}


//Render the cell grid
void Automata::RenderCells(int left, int top, SDL_Surface* tex, SDL_Surface* surface) const
{
	SDL_Rect src;
	src.x = 0;
	src.y = 0;
	src.w = tex->w;
	src.h = tex->h;
	SDL_Rect dest;
	dest.x = 0;
	dest.y = 0;
	dest.w = tex->w;
	dest.h = tex->h;

	for (int d = 0; d < bufferDepth; d++)
		for (int c = 0; c < bufferWidth; c++)
			if (cellBuffer[d][c])
			{
				dest.x = left + c * (tex->w + 1);
				dest.y = top + d * (tex->h + 1);
				SDL_BlitSurface(tex, &src, surface, &dest);
			}
}


//Read out the value of this automata in binary
int Automata::ParseValue()
{
	int val = 0;
	for (int c = bufferWidth - 1; c >= 0; c--)
		if (cellBuffer[bufferDepth - 1][c]) val += 1<<c;
	return val; 
}


//Read out the value of the highest-order bit in this automata
int Automata::HighBit()
{
	for (int c = 0; c < bufferWidth; c++)
		if (cellBuffer[bufferDepth - 1][c])
			return 1<<(bufferWidth - c - 1);
	return 0;
}