#include "SDL.h"
#include <assert.h>
#include <time.h>
#include "Automata.h"
#include <stdlib.h>
#include "Notes.h"
#include "Synth.h"

/*
Generative Music using Cellular Automata
Tech Demo

by Brian MacIntosh

V0.5: 12/10/12

You are free to modify and distribute this code or binaries for commercial
or non-commercial purposes, provided this header remains intact.
No warranty of fitness for any purpose, express or implied, is given.
*/

#define MIN(a, b) ((a < b) ? a : b)

//Display mode settings
int screenWidth = 640;
int screenHeight = 480;
int screenDepth = 32;

float targetFrameRate = 60;

bool gameRunning = true;
bool paused = false;

//Screen drawing surface
SDL_Surface* surface;
SDL_Surface* cell;

Automata* chord;
Automata* selector;
Automata* rhythm;

int bpm = 60;
double secPerBeat;
double currentBeat = 0;

int beatCounter;
int beatsToTrans = 24;

int chordCount = 4;
int chordSize = 5;

SDL_AudioSpec audioFormat;

Synth* synth;

double** chords;


void initialize()
{
	srand(time(NULL));

	secPerBeat = 60.0 / bpm;

	cell = SDL_LoadBMP("cell.bmp");

	synth = new Synth();

	chord = new Automata(5, 20);
	chord->SetRandomlineChance(15);
	chord->SetEffectiveWidth(4);
	chord->SetRule(110); //41, 105, 154 also cool

	selector = new Automata(4, 20);
	selector->SetRandomlineChance(10);
	selector->SetRule(36); //153

	rhythm = new Automata(5, 20);
	//rhythm->SetRandomruleChance(16);
	rhythm->SetRandomlineChance(9);
	rhythm->SetRule(73); //57?, 82

	chords = (double**)calloc(chordCount, sizeof(double*));
	for (int c = 0; c < chordCount; c++)
		chords[c] = (double*)calloc(chordSize, sizeof(double));

	//Star Wars
	chords[0][0] = D3;
	chords[0][1] = G3;
	chords[0][2] = B3;
	chords[0][3] = D4;
	chords[0][4] = D2;

	chords[1][0] = G3;
	chords[1][1] = D3;
	chords[1][2] = B3;
	chords[1][3] = D4;
	chords[1][4] = G2;

	chords[2][0] = Fs3;
	chords[2][1] = D3;
	chords[2][2] = B3;
	chords[2][3] = D4;
	chords[2][4] = Fs2;

	chords[3][0] = E3;
	chords[3][1] = G3;
	chords[3][2] = C4;
	chords[3][3] = G2;
	chords[3][4] = E2;

	//Pirates
	/*chords[0][0] = A2;
	chords[0][1] = A3;
	chords[0][2] = A3;
	chords[0][3] = E3;
	chords[0][4] = G3;

	chords[1][0] = A2;
	chords[1][1] = A3;
	chords[1][2] = A3;
	chords[1][3] = Cs3;
	chords[1][4] = A4;

	chords[2][0] = A2;
	chords[2][1] = A3;
	chords[2][2] = B3;
	chords[2][3] = D4;
	chords[2][4] = Fs2;

	chords[3][0] = A2;
	chords[3][1] = A3;
	chords[3][2] = C3;
	chords[3][3] = G3;
	chords[3][4] = C4;

	chords[4][0] = A2;
	chords[4][1] = A3;
	chords[4][2] = A3;
	chords[4][3] = F3;
	chords[4][4] = A4;

	chords[5][0] = A2;
	chords[5][1] = A3;
	chords[5][2] = Cs3;
	chords[5][3] = E3;
	chords[5][4] = Cs4;*/
}


void update(double elapsed)
{
	if (paused) return;

	currentBeat += elapsed;

	int beat = rhythm->HighBit();
	if (beat < 2) beat = 2;
	if (beat > 8) beat = 8;
	
	double beatLen = secPerBeat * 2 / beat;
	if (currentBeat > beatLen)
	{
		currentBeat -= beatLen;
		beatCounter++;

		chord->Cycle();
		selector->Cycle();
		rhythm->Cycle();

		//Send tones to a synth
		beat = rhythm->HighBit();
		if (beat < 2) beat = 2;
		if (beat > 8) beat = 8;
		beatLen = secPerBeat * 2 / beat;
		selector->RenderSynth(synth, beatLen * 1000);

		int b = selector->ParseValue();
		for (int c = 0; c < chordCount; c++)
			chord->SetFreq(c, chords[selector->ParseValue() % chordCount][c]);

		//Random rule!
		if (beatCounter >= beatsToTrans)
		{
			beatCounter = 0;
			selector->SetRule(rand() % 256);
			//rhythm->SetRule(rand() % 256);
		}
	}
}


void renderAudio(void *unused, Uint8 *stream, int len)
{
	chord->RenderSines(stream, len, audioFormat);
	//synth->Render(stream, len, audioFormat);
}


void draw()
{
	SDL_FillRect(surface, NULL, 0);

	chord->RenderCells(50, 50, cell, surface);
	selector->RenderCells(250, 50, cell, surface);
	rhythm->RenderCells(400, 50, cell, surface);
}


/*
Entry point
*/
int main(int argc, char* argv[])
{
	//Load SDL
	if(SDL_Init(SDL_INIT_EVERYTHING) < 0) return 1;
	surface = SDL_SetVideoMode(
		screenWidth, screenHeight, screenDepth,SDL_DOUBLEBUF);
	assert(surface != NULL);

	//Config windows
	SDL_WM_SetCaption("SDL", "");

	//Config audio
	audioFormat.freq = 22050;
    audioFormat.format = AUDIO_S16;
    audioFormat.channels = 1;
    audioFormat.samples = 512;
    audioFormat.callback = renderAudio;
    audioFormat.userdata = NULL;
	assert(SDL_OpenAudio(&audioFormat, NULL) >= 0);

	//Initialize loop
	clock_t ticksPerFrame = CLOCKS_PER_SEC / targetFrameRate;
	clock_t lastUpdate = clock();
	clock_t lastDraw = 0;

	//Load game
	initialize();
	SDL_PauseAudio(0);

	//Game loop
	while (gameRunning)
	{
		//Handle events
		SDL_Event ev;
		while (SDL_PollEvent(&ev))
		{
			if (ev.type == SDL_QUIT)
			{
				gameRunning = false;
			}
			else if (ev.type == SDL_KEYDOWN && ev.key.keysym.sym == SDLK_SPACE)
				paused = !paused;
		}

		//Update logic
		clock_t thisUpdate = lastUpdate;
		lastUpdate = clock();
		update((lastUpdate - thisUpdate) / (1.0 * CLOCKS_PER_SEC));

		//Draw to screen
		if (targetFrameRate == -1 || clock() - lastDraw >= ticksPerFrame)
		{
			lastDraw = clock();
			draw();
			SDL_Flip(surface);
		}
	}

	SDL_CloseAudio();
	SDL_Quit();

	free(chord);
	free(selector);
	free(rhythm);

	return 0;
}